#!/bin/python
##############################################################
#    Yasuko Matsubara 
#    Date: 11-10-2014
##############################################################
import sys
import math as M
from scipy.stats import poisson
import matplotlib.pyplot as plt
import pylab 
import numpy as np
import tool
import exo
INF=10000000000
XMIN=0; XMAX=10;
PP=52


#####################
# Given: 
#    n: duration 
#    d: dimension
#    X0[d]: initial values
#    r[d]: growth rate
#    K[d]: carrying capacity
#    A[d][d]: interaction matrix 
# -------------------
# Output: X[d][n]
#####################
def _EcoWeb_Base(model, n):
    d=model.d; #n=model.n;
    X0 = model.X0 
    r=model.r 
    K=model.K
    A=model.A
    X  = np.zeros((d,n)) 
    #--- init value ---#
    for i in range(0,d):
        X[i][0]=X0[i]
    #--- start (t=0,...n-1) ---#
    for t in range(0,n-1):
        for i in range(0,d):
            cum = 0.0;
            for j in range(0,d):
                cum += A[i][j]*X[j][t]
            sums=1.0 - cum/K[i];
            X[i][t+1] = X[i][t] + X[i][t]*r[i] * sums
            if(X[i][t+1]<XMIN or X[i][t+1]>XMAX):
                X[i][t+1] = 0;
    #--- return variables ---#
    return X

def _EcoWeb_Full(model, n):
    d = model.d
    # ecoweb-base
    X = _EcoWeb_Base(model, n)
    E = model.exo.getE(n)+1.0
    # ecoweb-full
    Y = np.zeros((d, n))
    for i in range(0,d):
        for t in range(0,n):
            Y[i][t] = X[i][t]*E[i][t]
    return Y

#----------------------------------#
#
#     class EcoWeb
#
#----------------------------------#
class EcoWeb:
    # d: dimension 
    # n: duration
    # X0[d]:   initial values
    # r[d]:    growth rate
    # A[d][d]: interaction matrix 
    # K[d]: carrying capacity
    # X0_V=[]; r_V=[]; A_V=[]; K_V=[];
    # model.exo = exo.EXO(d,n,Pp=PP)

    def __init__(self, d, n, data, mtype):
        self.mtype=mtype
        self.data=data
        self.allocModel(d, n)
        _initParams(self)
    def initParams(self):
        _initParams(self)
    def initVary(self, wantVary):
        _initVary(self, wantVary)
    def allocModel(self, d, n):
        _allocModel(self, d, n)
    def copyModel(self):
        return _copyModel(self)

    def getModel(self, n=-1):
        # default: generate X of length n
        if(n==-1): n = self.n 
        # basic model (without exo)
        if(self.mtype=='k'):
            return _EcoWeb_Base(self, n)
        # with user activities
        elif(self.mtype=='u'):
            return _EcoWeb_Full(self, n) 
        else:
            tool.error('usage: u or k')
        return []

    def dist(self, scale):
        return _dist(self, scale)
    def dist_each(self, scale):
        return _dist_each(self, scale)

    def saveModel(self, outdir, wantBlock=False):
        if(outdir!=''):
            outdir=(outdir+self.mtype+'_')
        tool.comment("save data: %s"%outdir)
        # fitting results (png)
        _plotFit(  self, outdir, wantBlock, 2)
        _plotModel(self, outdir, wantBlock, 1)
        # parameter set
        #self.printParams('');
        self.printParams(outdir)
        if(self.mtype=='u'):
            self.exo.printParams(outdir)
 
    def printParams(self, fn):
        _printParams(self, fn)
    def readParams(self, fn):
        _readParams(self, fn)
        if(self.mtype=='u'):
            self.exo.readParams(fn)


#----------------------------------#
#
#     private functions 
#
#----------------------------------#
def _allocModel(model, d, n):
    model.d  = d # of dimension
    model.n  = n # of duration
    model.X0 = np.zeros(d)    # X0: initial values
    model.r  = np.zeros(d)    # r: growth rate
    model.A  = np.identity(d) # A: interaction matrix
    model.K  = np.ones(d)     # K: carrying capacity
    # vary (1: true, 0: false)
    model.initVary(1)
    # user activity
    model.exo = exo.EXO(d,n,Pp=PP)

def _initParams(model):
    for i in range(0,model.d):
        model.K[i] = 1.0
        model.X0[i]= 0.5
        model.r[i] = 0.01 
        for j in range(0,model.d):
            model.A[i][j]= 0.1 
        model.A[i][i]=1.0

def _initVary(model, wantVary):
    # vary (1: true, 0: false)
    d = model.d
    model.X0_V  = wantVary*np.ones(d)
    model.r_V   = wantVary*np.ones(d)
    model.A_V   = wantVary*np.ones((d,d))
    model.K_V   = wantVary*np.ones(d)

def _copyModel(model):
    new = EcoWeb(model.d, model.n, model.data, model.mtype)
    new.X0 = model.X0.copy()
    new.r  = model.r.copy()
    new.A  = model.A.copy()
    new.K  = model.K.copy()
    new.initVary(1)
    new.exo = exo.EXO(new.d,new.n,Pp=PP)
    new.exo.optk=model.exo.optk
    new.exo.W = model.exo.W.copy()
    new.exo.B = model.exo.B.copy()
    return new

def _readParams(model, fn):
    print "read base model parameters (X0, r, a) ..."
    f = open(fn, 'r')#
    d = int(f.readline().rstrip().split(' ')[1])
    n = int(f.readline().rstrip().split(' ')[1])
    if(model.d < d):
        tool.warning('d(current) < d(model)')
        model.d = d
    if(model.n < n):
        tool.warning('n(current) < n(model)')
        model.n = n
    model.allocModel(model.d,model.n)
    # X0
    f.readline()
    X0 = f.readline().rstrip().split(' ')
    for i in range(0,d):
        model.X0[i] = float(X0[i])
    # r
    f.readline()
    r = f.readline().rstrip().split(' ')
    for i in range(0,d):
        model.r[i] = float(r[i])
    # K
    f.readline()
    K = f.readline().rstrip().split(' ')
    for i in range(0,d):
        model.K[i] = float(K[i])
    # a
    f.readline()
    for i in range(0,d):
        a = f.readline().rstrip().split(' ')
        for j in range(0,d):
            model.A[i][j] = float(a[j])
    f.close()
    model.printParams('')

def _printParams(model, fn):
    if(fn!=''):
        f = open(fn+'model', 'w')
        sys.stdout = f
    print "d= %d" %model.d
    print "n= %d" %model.n
    print "X0:"
    for i in range(0,model.d):
        print u"%.4e" % model.X0[i],
    print ""
    print "r:"
    for i in range(0,model.d):
        print u"%.8f" % model.r[i],
    print ""
    print "K:"
    for i in range(0,model.d):
        print u"%.8f" % model.K[i],
    print ""
    print "A:"
    for i in range(0,model.d):
        for j in range(0,model.d):
            print u"%.8f" % model.A[i][j],        
        print ""
    if(fn!=''):
        sys.stdout = sys.__stdout__
        f.close()

def _dist(model, scale):
    data= model.data
    if(data==[]): return 0
    (d,n)=np.shape(data)
    X = model.getModel();
    X = X[0:d,0:n]
    a = data.flatten()
    b = X.flatten()
    return np.sqrt(np.mean(pow(a-b, 2)))

def _dist_each(model, scale):
    data= model.data
    if(data==[]): return 0
    (d,n)=np.shape(data)
    X = model.getModel();
    X = X[0:d,0:n]
    D = np.zeros(d);
    for i in range(0,d):
        D[i] = np.sqrt(np.mean(pow(data[i]-X[i], 2)))
    return D

def _plotFit(model, fn, wantBlock, figid):
    tool.figure(figid)
    X = model.getModel();
    # (a) plot sequences
    pylab.title("rse=%f"%model.dist('lin'))
    pylab.xlabel("Time")
    pylab.ylabel("x_i @ time ")
    # (a-1) plot original data
    pylab.gca().set_color_cycle(None)
    if(model.data!=[]):
        pylab.plot(model.data.T, '-', lw=0.2)
    # (a-2) plot model fitting result
    pylab.gca().set_color_cycle(None)
    pylab.plot(X.T, lw=1)
    plt.xlim(0,model.n)
    pylab.draw()
    pylab.show(block=wantBlock)
    # (b) save data
    if(fn!=''):
        pylab.savefig(fn+'fit.png')
        pylab.savetxt(fn+'seq_fit.txt', model.getModel().T, delimiter=" ", fmt="%.16f")
        if(model.data!=[]):
            pylab.savetxt(fn+'seq_org.txt', model.data.T, delimiter=" ", fmt="%.16f")

def _plotModel(model, fn, wantBlock, figid):
    tool.figure(figid)
    d = model.d; n=model.n;
    vmA=max(abs(model.A.flatten()))
    vmr=max(abs(model.r.flatten()))
    vmXi=max(abs(model.X0.flatten()))
    vmK=max(abs(model.K.flatten()))
    # (a-1) X0
    plt.subplot(261)
    plt.imshow(model.X0.reshape(d,1), interpolation='nearest',vmin=-vmXi,vmax=vmXi)
    #plt.gray()
    pylab.title("X0")
    plt.xticks([]); plt.yticks(np.arange(0,d), range(1,d+1))
    # (a-2) original r
    plt.subplot(262)
    plt.imshow(model.r.reshape(d,1), interpolation='nearest', vmin=-vmr,vmax=vmr)
    pylab.title("r")
    plt.xticks([]); plt.yticks(np.arange(0,d), range(1,d+1))
    # (a-3) original K
    plt.subplot(263)
    plt.imshow(model.K.reshape(d,1), interpolation='nearest', vmin=-vmK,vmax=vmK)
    pylab.title("K")
    plt.xticks([]); plt.yticks(np.arange(0,d), range(1,d+1))
    # (b-1) original A
    plt.subplot(222)
    plt.imshow(model.A, interpolation='nearest',vmin=-vmA,vmax=vmA)
    plt.xticks(np.arange(0,d), range(1,d+1))
    plt.yticks(np.arange(0,d), range(1,d+1))
    #plt.colorbar()
    pylab.title("A")
    # (b-2) normalized A
    An = tool.normalizeM(model.A, d, d)
    # (c) E matrix
    pylab.subplot(223)
    E = model.exo.getE(n)
    #E = tool.normalizeM(E, d, n)
    pylab.plot(E.T)
    plt.xlim(0,model.exo.Pp)
    plt.ylim(min(E.flatten())-0.01, max(E.flatten())+0.01)
    pylab.title("E")
    pylab.subplot(224)
    if(model.exo.optk>0):
        pylab.plot(model.exo.B[0:model.exo.optk,:].T)
        plt.xlim(0,model.exo.Pp)
    pylab.title("B (k=%d)"%model.exo.optk)
    # END
    pylab.draw()
    #pylab.show(block=wantBlock)
    # save data if you want
    if(fn!=''):
        pylab.savefig(fn+'A.png')


#---------------#
#     main      #
#---------------#
if __name__ == "__main__":
    # basic parameters
    n  = 500;
    d  = 2 #3;
    dat=[]
    model = EcoWeb(d, n, dat, 'u');
    #model.exo.B[0]=pylab.loadtxt("E")
    model.exo.B[0]=np.random.rand(PP)
    model.exo.W[0][0]=0.1
    model.exo.W[1][0]=0.5
    model.exo.B[0]=dat
    model.exo.optk=1
    model.printParams('');


    #model.readParams('../output/synthetic/syn1/u_model')

    _plotFit(model, '', 'True', 1);





