import pylab
import sys
import numpy as np
import matplotlib.pyplot as plt
import tool
import EcoWeb 
import time
import ica
import os

try:
    from lmfit import Parameters, Minimizer
except:
    tool.error("can not find lmfit - please see http://lmfit.github.io/lmfit-py/")
# basic vals
INF=1.e+8
ZERO=1.e-8 
YES=1; NO=0;
# for lmfit 
XTL=1.e-8
FTL=1.e-8
# debug, etc. 
DISPLAY=NO #(YES/NO)
COMMENT=NO #(YES/NO)
# for StepFit()
MAXITER=4
MAXFEV=10000 
MAXFEV_m=100
DELTA=1.e-2
# fitting (RMSE or RMSE(sqrt()))
BASEFIT_SQRT=NO
EXOFIT_SQRT=YES
# with/without ica_opt
OPTK=YES
BFIT=YES
SEED=1
RANDS=NO # random seed (YES/NO)
#--------------------#
#    model fitting   #
#--------------------#
def LMfit(model, outdir, type, wantDisp):
    # set seed (optional)
    if(RANDS): global SEED; SEED=np.random.seed(os.getpid());
    if(wantDisp): global DISPLAY; DISPLAY=YES;
    tool.comment("start LM fitting") 
    #---setting ---------------------#
    #---  do fit, here with leastsq model ---#
    tic = time.clock()
    if(type=='naive'):
        global OPTK; OPTK=NO  # without k-optimization
        global BFIT; BFIT=NO  # without bounding
        model=_StepFit(model, MAXFEV, YES, model.d)
    if(type=='proposed'):
        OPTK=YES # with k-optimization
        model=_EcoWeb_Fit(model, MAXFEV_m)
    toc = time.clock(); fittime= toc-tic;
    #--------------------------------#
    #--- show results and save data ---#
    tool.comment("end LM fitting")
    print "time: %f"%fittime
    print "rmse: %f"%model.dist('lin')
    if(outdir!=''):
        model.saveModel(outdir+type+'_') 
    #--------------------------------#
    return model

def _StepFit(model, maxiter, wantInit, maxk):
    err_prev=INF; model_prev=[]; 
    #tool.comment("# of iter: %s"%ITER)
    if(wantInit): _initParams(model)
    for i in range(0,MAXITER):
        # (a) base fitting
        _LMfit_base(model, maxiter)
        # (b) exo (seasonal)  fitting
        _LMfit_exo(model, maxiter, maxk);
        # (a) if diff < DELTA, stop iter
        err=model.dist('lin')
        if(DISPLAY): model.saveModel('')
        if(COMMENT): print "StepFit: (err:%.2e)"%err
        if(abs(err-err_prev)<DELTA):
            if(COMMENT): print '# of opt-iter:%d (err:%.2e)'%((i+1),err); break;
        if(err_prev < err): 
            model=model_prev; err=err_prev; 
        err_prev=err; model_prev=model;
    return model



def _copyBaseParams(mfrom, ifrom, mto, ito):
    mto.X0[ito]=mfrom.X0[ifrom]
    mto.K[ito]=mfrom.K[ifrom]
    mto.r[ito]=mfrom.r[ifrom]
    mto.A[ito][ito]=mfrom.A[ifrom][ifrom]


def _EcoWeb_Fit(modelG, maxiter):
    d = modelG.d; modelG.mtype='u'
    # reset A matrix (all zero)
    modelG.A=1.e-20*np.ones((d,d))
    # === (I) solo-fitting === #
    tool.comment('(I) individual fitting')
    d_p=1; maxk=1;
    for i in range(0,d):
        data_p=modelG.data[[i],:]
        model_p=EcoWeb.EcoWeb(d_p,modelG.n,data_p, 'u')
        _StepFit(model_p, maxiter, YES, maxk)
        # assign params modelG <- model_p 
        _copyBaseParams(model_p, 0, modelG, i)
        modelG.exo.B[i]    = model_p.exo.B[0]
        modelG.exo.W[i][i] = model_p.exo.W[0]
    modelG.exo.optk = d
    # === (II) pair-fitting === #
    tool.comment('(II) pair fitting')
    d_p=2; maxk=1; bestk=d;
    V_bi=np.ones(d);
    for iter in (0,d*d):
        errG_prev=modelG.dist('lin')
        if(DISPLAY): modelG.saveModel('') 
        err_b=INF; model_b=[]; i_b=-1; j_b=-1; # best pair
        Dist=modelG.dist_each('lin')
        # (a-1) choose single seq i
        i=np.argmax(Dist) 
        if(COMMENT): print "choose single seq: %d (err:%.2e)"%(i,Dist[i])
        # (a-2) if fitting accuracy is good enough, stop fitting
        if(Dist[i] < DELTA): break
        # (b) pair fit with seq j
        for j in range(0,d):
            if(i==j or V_bi[i]!=1 or V_bi[j]!=1): 
                continue
            data_p=modelG.data[[i,j],:]
            model_p=EcoWeb.EcoWeb(d_p,modelG.n,data_p, 'u')
            if(NO):
                _StepFit(model_p, maxiter, YES, maxk)
            else:
                _copyBaseParams(modelG, i, model_p, 0)
                _copyBaseParams(modelG, j, model_p, 1)
                _StepFit(model_p, maxiter, NO, maxk)

            err_p=model_p.dist('lin')
            if(COMMENT): print "-pair fitting: (%d,%d), (err:%.2e)"%(i,j,err_p) 
            if(err_b > err_p): # update best pair
                model_b=model_p; i_b=i; j_b=j; err_b=err_p; #err_diff;
        if(COMMENT): print "--best pair: (%d,%d), (err:%.2e)"%(i_b,j_b,err_b)
        # (c) if no candidate, then stop pair fitting
        if(i_b==-1 or j_b==-1): break;
        V_bi[i_b]=0; V_bi[j_b]=0; bestk=bestk-1; # fix pair 
        _copyBaseParams(model_b, 0, modelG, i_b)
        _copyBaseParams(model_b, 1, modelG, j_b)
        modelG.A[i_b][j_b]=model_b.A[0][1]; modelG.A[j_b][i_b]=model_b.A[1][0]
        _LMfit_exo(modelG, maxiter, bestk)    
        errG = modelG.dist('lin')
        if(DISPLAY): modelG.saveModel('')
    # === (III) multi-fitting === #
    tool.comment('(III) full fitting')
    modelG.initVary(1)
    _StepFit(modelG, maxiter, NO, bestk)
    return modelG

#--------------------------------------#
#    objective functions  #
#    return the array to be minimized
#--------------------------------------#
def _distfunc(params, model, outdir):
    tool.dotting('.')
    d = model.d; data = model.data;
    for i in range(0,d):
        model.K[i]   = abs(params['K_%i'%i].value);
        model.X0[i]  = abs(params['X0_%i'%i].value);
        model.r[i]   = abs(params['r_%i'%i].value);
        for j in range(0,d):
            model.A[i][j] = abs(params['A_%i_%i'%(i,j)].value);
    _bounding(model)
    # make model
    (X) = model.getModel();
    if(BASEFIT_SQRT): data=np.sqrt(data); X=np.sqrt(X);
    return pow(data.flatten() - X.flatten(), 2)

def _distfunc_exo(params, model, outdir):
    tool.dotting('o')
    d = model.d; data = model.data;
    for k in range(0,model.exo.optk):
        for i in range(0,d):
            model.exo.W[i][k] = params['W_%i_%i'%(i,k)].value;
    # make model
    (X) = model.getModel();
    if(EXOFIT_SQRT): data=np.sqrt(data); X=np.sqrt(X);
    return pow(data.flatten() - X.flatten(), 2)


def _bounding(model):
    if(BFIT==NO): return; 
    #--- L/U bounding ---#
    rmin=0.001; Amax=1.0; Kmax=1.2; Kmin=0.01;
    for i in range(0,model.d):
        if(model.r_V[i]==1 and model.r[i]<rmin): model.r[i] = rmin;
        if(model.K_V[i]==1 and model.K[i]>Kmax): model.K[i] = Kmax; 
        if(model.K_V[i]==1 and model.K[i]<Kmin): model.K[i] = Kmin; 
        for j in range(0,model.d):
            if(model.A_V[i][j]==1 and abs(model.A[i][j]>Amax)): model.A[i][j]=Amax;

def _initParams(model):
    model.initParams()
    r=0.01*np.random.random()
    for i in range(0,model.d):
        model.K[i]  = max(model.data[i])+r
        model.X0[i] = model.data[i][0]+0.1+r
        model.r[i]  = 0.05+0.01*r
        for j in range(0,model.d):
            model.A[i][j] = 0.1+r
        model.A[i][i] = 1.0

#---------------------------#
# base fitting (X0, r, K, A)
#---------------------------#
def _LMfit_base(model, maxiter):
    #--- (a) create a set of Parameters -----#
    params = Parameters()
    for i in range(0,model.d):
        params.add('K_%i'%i, value=model.K[i], min=-10, max=10, vary=model.K_V[i])
        params.add('X0_%i'%i,value=model.X0[i],min=-10, max=10, vary=model.X0_V[i])
        params.add('r_%i'%i, value=model.r[i], min=-10, max=10,vary=model.r_V[i])
        for j in range(0,model.d):
            params.add('A_%i_%i'%(i,j),value=model.A[i][j],min=-10,max=10,vary=model.A_V[i][j])
        params['A_%i_%i'%(i,i)].vary=False
    #--- (b) start fitting ------------------#
    lmsol = Minimizer(_distfunc, params, fcn_args=(model, ''))
    lmsol.leastsq(xtol=XTL, ftol=FTL, maxfev=maxiter)
    print ""

#---------------------------#
# seasonal fitting (W, B)
#---------------------------#
def _LMfit_exo(model, maxiter, maxk):
    if(model.mtype!='u'): return;
    d = model.d; n = model.n; data = model.data;
    if(maxk<1): return; 
    #--- (I) make exo components ---#
    # (a-1) X: model without exo pattern
    model.mtype='k'
    X = model.getModel();
    model.mtype='u'
    # (a-2) make diff patterns ---#
    delta=1.0 # to avoid 1/0 fitting
    diff=(data.T-(X.T))/(X.T+delta); diff=diff.T 
    Pp = model.exo.Pp;   # Pp: period
    s = int(n/Pp);       #  s: #ofsubseqs 
    V=np.zeros((d*s,Pp)) #  V: subseq set
    # (a-3) create subseq matrix V
    for i in range(0,d):
        for t in range(0,Pp*s): #range(0,n):
            V[int(i*s+t/Pp)][np.mod(t,Pp)]=diff[i][t]
    # (b-1) find k components using ICA
    if(OPTK): mink=1  # 1<=optk<=maxk
    else: mink=maxk   # optk=maxk (naive method)
    (optk, Stmp, Atmp, Mtmp) = ica.ica_optk(V.T, mink, maxk)
    #(Stmp, Atmp, Mtmp) = ica.ica(V.T, maxk); optk=maxk; #if not opt 
    if(COMMENT): print 'ica_opt: optk=%d (maxk=%d)'%(optk, maxk) 
    (Stmp, Atmp, Mtmp) = ica.positive(Stmp, Atmp, Mtmp)
    # (b-2) update exo parameters (k, B)
    model.exo.resetEXO(); model.exo.optk=optk;
    Stmp=Stmp.T; model.exo.B[0:optk]=Stmp[0:optk,:];
    # normalize B (currently, not required)
    #model.exo.B=tool.normalize01(model.exo.B+1, optk, Pp)-1.0
    #--- (II) exo W witting ---#
    params = Parameters()
    for k in range(0,model.exo.optk):
        for i in range(0,d):
            params.add('W_%i_%i'%(i,k), value=model.exo.W[i][k]) 
    #--- start fitting -----#
    lmsol = Minimizer(_distfunc_exo, params, fcn_args=(model,''))
    lmsol.leastsq(xtol=XTL, ftol=FTL, maxfev=maxiter)
    print ""


